use crate::db::{Connection, StatementContext, StatementRow};
use crate::schema::entity::helpers::check_assoc;
use crate::schema::{AssocData, IDWrap, LocalSide};
use crate::{
    schema::datum::{Datum, DatumList},
    schema::entity::{Entity, EntityID, EntityPart, EntityPartList, EntityPartVisitor},
};
use crate::{DBResult, Error};
use std::collections::HashMap;
use std::hash::{Hash, Hasher};

pub(crate) mod components;

pub(crate) fn insert<E: Entity>(conn: &Connection, value: E) -> DBResult<E::ID> {
    struct InsertQuery<E: Entity>(std::marker::PhantomData<E>);

    conn.with_prepared(
        std::any::TypeId::of::<InsertQuery<E>>(),
        || {
            let mut part_names = String::new();
            let mut placeholders = String::new();
            struct PartNameVisitor<'a>(&'a mut String, &'a mut String);
            impl<'a> EntityPartVisitor for PartNameVisitor<'a> {
                fn visit<EP: EntityPart>(&mut self) {
                    // if this is a set-association, then we don't actually want to do anything
                    // with it here; it doesn't have a column
                    if check_assoc::<EP>() {
                        return;
                    }

                    if !self.0.is_empty() {
                        self.0.push_str(", ");
                        self.1.push_str(", ");
                    }
                    self.0.push('`');
                    self.0.push_str(EP::part_name());
                    self.0.push('`');
                    self.1.push('?');
                }
            }

            E::accept_part_visitor(&mut PartNameVisitor(&mut part_names, &mut placeholders));

            format!(
                "INSERT INTO `{}` ({}) VALUES ({}) RETURNING `id`",
                E::entity_name(),
                part_names,
                placeholders
            )
        },
        |mut ctx| {
            struct PartBinder<'a, 'b>(&'a mut StatementContext<'b>, i32);
            impl<'a, 'b> EntityPartVisitor for PartBinder<'a, 'b> {
                fn visit_datum<EP: EntityPart>(&mut self, datum: &EP::Datum) {
                    // skip associations, as with the query preparation above
                    if check_assoc::<EP>() {
                        return;
                    }

                    datum.bind_to(self.0, self.1);
                    self.1 += 1;
                }
            }

            value.accept_part_visitor_ref(&mut PartBinder(&mut ctx, 1));

            ctx.run()?
                .ok_or(Error::InternalError("No result row from INSERT query"))
                .map(|r| <E::ID>::from_raw(r.read(0).expect("couldn't read resulting ID")))
        },
    )
}

#[derive(Debug, Clone, Copy, Hash, PartialEq, Eq)]
pub(crate) enum QueryPart {
    Root,
    Columns,
    From,
    Set,
    Join,
    Where,
    Trailing,
}

#[derive(Debug)]
pub struct Query {
    parts: HashMap<QueryPart, Vec<String>>,
}

impl Query {
    pub(crate) fn new() -> Self {
        Self {
            parts: Default::default(),
        }
    }

    pub(crate) fn attach(mut self, qp: QueryPart, val: String) -> Self {
        self.attach_mut(qp, val);
        self
    }

    pub(crate) fn replace(mut self, qp: QueryPart, val: String) -> Self {
        self.parts.remove(&qp);
        self.attach(qp, val)
    }

    pub(crate) fn attach_mut(&mut self, qp: QueryPart, val: String) {
        self.parts.entry(qp).or_default().push(val);
    }

    pub(crate) fn assemble(mut self) -> String {
        let root = self.parts.remove(&QueryPart::Root).unwrap().remove(0);

        let columns_ = match self.parts.remove(&QueryPart::Columns) {
            None => String::new(),
            Some(v) => v
                .into_iter()
                .reduce(|a, b| format!("{}, {}", a, b))
                .unwrap(),
        };

        let from_ = match self.parts.remove(&QueryPart::From) {
            None => String::new(),
            Some(v) => {
                format!(
                    "FROM {}",
                    v.into_iter()
                        .reduce(|a, b| format!("{}, {}", a, b))
                        .unwrap()
                )
            }
        };

        let set_ = match self.parts.remove(&QueryPart::Set) {
            None => String::new(),
            Some(v) => {
                format!(
                    "SET {}",
                    v.into_iter()
                        .reduce(|a, b| format!("{}, {}", a, b))
                        .unwrap()
                )
            }
        };

        let join_ = match self.parts.remove(&QueryPart::Join) {
            None => String::new(),
            Some(v) => v
                .into_iter()
                .map(|j| format!("INNER JOIN {}", j))
                .reduce(|a, b| format!("{} {}", a, b))
                .unwrap(),
        };

        let where_ = match self.parts.remove(&QueryPart::Where) {
            None => String::new(),
            Some(v) => {
                format!(
                    "WHERE {}",
                    v.into_iter()
                        .reduce(|a, b| format!("{} AND {}", a, b))
                        .unwrap()
                )
            }
        };

        let trailing_ = match self.parts.remove(&QueryPart::Trailing) {
            None => String::new(),
            Some(v) => v.into_iter().reduce(|a, b| format!("{} {}", a, b)).unwrap(),
        };

        /*println!(
            "built SQL query: {} {} {} {} {} {} {}",
            root, columns_, from_, set_, join_, where_, trailing_
        );*/
        // log::trace!("built SQL query: {} {} {}", root, set_, where_);

        format!(
            "{} {} {} {} {} {} {}",
            root, columns_, from_, set_, join_, where_, trailing_
        )
    }
}

pub(crate) struct AssocNames {
    local_name: &'static str,
    remote_name: &'static str,
    part_name: &'static str,
    dist_name: &'static str,
    domain_name: &'static str,
    range_name: &'static str,
    local_field: &'static str,
    remote_field: &'static str,
}

impl AssocNames {
    fn collect<AI: AssocInterface>(iface: &AI) -> DBResult<AssocNames> {
        let adata = iface.get_data()?;
        let local_name = adata.local_name;
        let remote_name = <AI::RemoteEntity>::entity_name();
        let part_name = adata.part_name;
        let dist_name = iface.get_distinguishing_name()?;

        let (domain_name, range_name) = match AI::SIDE {
            LocalSide::Domain => (local_name, remote_name),
            LocalSide::Range => (remote_name, local_name),
        };
        let (local_field, remote_field) = match AI::SIDE {
            LocalSide::Domain => ("domain", "range"),
            LocalSide::Range => ("range", "domain"),
        };
        Ok(Self {
            local_name,
            remote_name,
            part_name,
            dist_name,
            domain_name,
            range_name,
            local_field,
            remote_field,
        })
    }

    fn assoc_name(&self) -> String {
        format!(
            "{domain_name}_{range_name}_assoc_{dist_name}",
            domain_name = self.domain_name,
            range_name = self.range_name,
            dist_name = self.dist_name
        )
    }
}

fn hash_of<T: Hash>(val: T) -> u64 {
    let mut hasher = std::collections::hash_map::DefaultHasher::new();
    val.hash(&mut hasher);
    hasher.finish()
}

pub trait AssocInterface: 'static {
    type RemoteEntity: Entity;
    fn get_data(&self) -> DBResult<&AssocData>;
    fn get_distinguishing_name(&self) -> DBResult<&'static str>;
    const SIDE: LocalSide;

    fn connect_to(&self, remote_id: <Self::RemoteEntity as Entity>::ID) -> DBResult<()>
    where
        Self: Sized,
    {
        let adata = self.get_data()?;
        let an = AssocNames::collect::<Self>(self)?;

        // second, add to the assoc table
        adata.conn.with_prepared(
            hash_of(("connect", an.local_name, an.remote_name, an.part_name)),
            || {
                format!(
                    "insert into `{assoc_name}` (`{local_field}`, `{remote_field}`) values (?, ?)",
                    assoc_name = an.assoc_name(),
                    local_field = an.local_field,
                    remote_field = an.remote_field
                )
            },
            |mut ctx| {
                ctx.bind(1, adata.local_id)?;
                ctx.bind(2, remote_id.into_raw())?;

                ctx.run().map(|_| ())
            },
        )
    }

    fn disconnect_from(&self, remote_id: <Self::RemoteEntity as Entity>::ID) -> DBResult<()>
    where
        Self: Sized,
    {
        let adata = self.get_data()?;
        let an = AssocNames::collect::<Self>(self)?;

        // second, add to the assoc table
        adata.conn.with_prepared(
            hash_of(("disconnect", an.local_name, an.remote_name, an.part_name)),
            || {
                format!(
                    "delete from `{assoc_name}` where `{local_field}` = ? and `{remote_field}` = ?",
                    assoc_name = an.assoc_name(),
                    local_field = an.local_field,
                    remote_field = an.remote_field
                )
            },
            |mut ctx| {
                ctx.bind(1, adata.local_id)?;
                ctx.bind(2, remote_id.into_raw())?;

                ctx.run().map(|_| ())
            },
        )
    }

    fn insert(&self, value: Self::RemoteEntity) -> DBResult<<Self::RemoteEntity as Entity>::ID>
    where
        Self: Sized,
    {
        // we're doing two things:
        // - inserting the entity into the target table
        // - adding the association row into the assoc table

        let adata = self.get_data()?;

        // so first, the remote table
        let remote_id = insert(&adata.conn, value)?;
        // then the association
        self.connect_to(remote_id)?;
        // TODO: handle error case of associate_with() fails but insert() succeeds
        Ok(remote_id)
    }
}

// ----------------------------------------------------------------------
// New query interface
// ----------------------------------------------------------------------

pub trait OutputContainer: 'static {
    fn assemble_from(conn: &Connection, stmt: StatementContext<'_>) -> DBResult<Self>
    where
        Self: Sized;
}

fn assemble_single<T: Entity>(conn: &Connection, row: &mut StatementRow) -> IDWrap<T> {
    let id = row.read::<i64>(0).expect("couldn't read ID");
    let datum_list = <T::Parts>::build_datum_list(conn, row).expect("couldn't build datum list");
    IDWrap::new(T::ID::from_raw(id), T::build(datum_list))
}

impl<T: Entity> OutputContainer for Option<IDWrap<T>> {
    fn assemble_from(conn: &Connection, ctx: StatementContext<'_>) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(ctx.run()?.map(|mut r| assemble_single(conn, &mut r)))
    }
}

impl<T: Entity> OutputContainer for Vec<IDWrap<T>> {
    fn assemble_from(conn: &Connection, ctx: StatementContext<'_>) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(ctx
            .iter()
            .map(|mut r| assemble_single(conn, &mut r))
            .collect())
    }
}

pub trait Queryable {
    type EntityOutput: Entity;
    type OutputContainer: OutputContainer;
    type StaticVersion: Queryable + 'static;

    fn build(&self) -> Query;
    fn bind(&self, stmt: &mut StatementContext, index: &mut i32);
    fn conn(&self) -> &Connection;

    // ----------------------------------------------------------------------
    // Verbs
    // ----------------------------------------------------------------------
    /// Count all entities in the current context.
    ///
    /// Returns the number of entities.
    fn count(self) -> DBResult<usize>
    where
        Self: Sized,
    {
        struct CountTag;
        self.conn().with_prepared(
            std::any::TypeId::of::<(Self::StaticVersion, CountTag)>(),
            || {
                self.build()
                    .replace(
                        QueryPart::Columns,
                        format!(
                            "COUNT(DISTINCT `{}`.`id`)",
                            Self::EntityOutput::entity_name()
                        ),
                    )
                    .assemble()
            },
            |mut ctx| {
                // starting index is 1
                let mut index = 1;
                self.bind(&mut ctx, &mut index);

                Ok(ctx
                    .run()?
                    .ok_or(Error::InternalError("no resulting rows from COUNT query"))?
                    .read::<i64>(0)? as usize)
            },
        )
    }
    /// Get all entities in the current context.
    fn get(self) -> DBResult<Self::OutputContainer>
    where
        Self: Sized,
    {
        struct GetTag;
        self.conn().with_prepared(
            std::any::TypeId::of::<(Self::StaticVersion, GetTag)>(),
            || self.build().assemble(),
            |mut ctx| {
                // starting index is 1
                let mut index = 1;
                self.bind(&mut ctx, &mut index);

                <Self::OutputContainer>::assemble_from(self.conn(), ctx)
            },
        )
    }
    /// Delete all entities in the current context.
    fn delete(self) -> DBResult<()>
    where
        Self: Sized,
    {
        struct DeleteTag;
        self.conn().with_prepared(
            std::any::TypeId::of::<(Self::StaticVersion, DeleteTag)>(),
            || {
                format!(
                    "DELETE FROM `{}` WHERE `id` = ({})",
                    Self::EntityOutput::entity_name(),
                    self.build()
                        .replace(
                            QueryPart::Columns,
                            format!("`{}`.`id`", Self::EntityOutput::entity_name())
                        )
                        .assemble()
                )
            },
            |mut ctx| {
                // starting index is 1
                let mut index = 1;
                self.bind(&mut ctx, &mut index);

                ctx.run()?;
                Ok(())
            },
        )
    }

    // ----------------------------------------------------------------------
    // Filtering methods
    // ----------------------------------------------------------------------
    /// Filter using the unique index on the entity.
    fn unique(
        self,
        values: <<<Self::EntityOutput as Entity>::Uniques as EntityPartList>::DatumList as DatumList>::Ref<'_>,
    ) -> impl Queryable<
        EntityOutput = Self::EntityOutput,
        OutputContainer = Option<IDWrap<Self::EntityOutput>>,
    >
    where
        Self: Sized,
    {
        components::UniqueComponent::new(self, values)
    }
    /// Filter using an arbitrary column on the entity.
    fn with<EP: EntityPart<Entity = Self::EntityOutput>>(
        self,
        part: EP,
        value: &EP::Datum,
    ) -> impl Queryable<EntityOutput = Self::EntityOutput, OutputContainer = Self::OutputContainer>
    where
        Self: Sized,
    {
        components::WithComponent::new(self, part, value)
    }

    /// Ask to return at most a single result
    fn first(
        self,
    ) -> impl Queryable<
        EntityOutput = Self::EntityOutput,
        OutputContainer = Option<IDWrap<Self::EntityOutput>>,
    >
    where
        Self: Sized,
    {
        components::SingleComponent::new(self)
    }

    // ----------------------------------------------------------------------
    // Association-following and joining methods
    // ----------------------------------------------------------------------
    /// Join based on an existing association
    fn join<AD: AssocInterface, EP: EntityPart<Entity = Self::EntityOutput, Datum = AD>>(
        self,
        part: EP,
    ) -> impl Queryable<EntityOutput = AD::RemoteEntity, OutputContainer = Vec<IDWrap<AD::RemoteEntity>>>
    where
        Self: Sized,
    {
        components::JoinComponent::<AD::RemoteEntity, Self::EntityOutput, _, Self>::new(self, part)
    }
}

// Generic implementation for all assoc specification types
impl<'a, AI: AssocInterface> Queryable for &'a AI {
    type EntityOutput = AI::RemoteEntity;
    type OutputContainer = Vec<IDWrap<AI::RemoteEntity>>;
    type StaticVersion = &'static AI;

    fn build(&self) -> Query {
        unreachable!()
    }

    fn bind(&self, _stmt: &mut StatementContext, _index: &mut i32) {
        unreachable!()
    }

    fn conn(&self) -> &Connection {
        &self.get_data().unwrap().conn
    }

    fn count(self) -> DBResult<usize> {
        components::AssocQueryable::new(self).count()
    }

    fn get(self) -> DBResult<Self::OutputContainer> {
        components::AssocQueryable::new(self).get()
    }

    fn delete(self) -> DBResult<()> {
        components::AssocQueryable::new(self).delete()
    }

    fn unique(
        self,
        values: <<<Self::EntityOutput as Entity>::Uniques as EntityPartList>::DatumList as DatumList>::Ref<'_>,
    ) -> impl Queryable<
        EntityOutput = Self::EntityOutput,
        OutputContainer = Option<IDWrap<Self::EntityOutput>>,
    > {
        components::AssocQueryable::new(self).unique(values)
    }

    fn with<EP: EntityPart<Entity = Self::EntityOutput>>(
        self,
        part: EP,
        value: &EP::Datum,
    ) -> impl Queryable<EntityOutput = Self::EntityOutput, OutputContainer = Self::OutputContainer>
    {
        components::AssocQueryable::new(self).with(part, value)
    }

    fn join<AD: AssocInterface, EP: EntityPart<Entity = Self::EntityOutput, Datum = AD>>(
        self,
        part: EP,
    ) -> impl Queryable<EntityOutput = AD::RemoteEntity, OutputContainer = Vec<IDWrap<AD::RemoteEntity>>>
    {
        components::AssocQueryable::new(self).join(part)
    }
}

#[cfg(test)]
mod query_build_test {
    #[test]
    fn simple_construction() {}
}
