use crate::{
    db::{StatementContext, StatementRow},
    schema::{AssocData, Datum},
    DBResult, Error,
};

impl Datum for time::OffsetDateTime {
    fn sql_type() -> &'static str {
        "text"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        let ts = self.unix_timestamp();
        ts.bind_to(stmt, index)
    }

    fn build_from(adata: AssocData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        let unix = i64::build_from(adata, stmt, index)?;
        Ok(Self::from_unix_timestamp(unix).map_err(|e| Error::UnknownValue(e.to_string()))?)
    }
}

impl Datum for String {
    fn sql_type() -> &'static str {
        "text"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        stmt.bind(index, self.as_str())
            .expect("couldn't bind string");
    }

    fn build_from(_: AssocData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        let val = stmt.read(*index)?;
        *index += 1;
        Ok(val)
    }
}

impl Datum for usize {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        stmt.bind(index, *self as i64)
            .expect("couldn't bind usize as i64");
    }

    fn build_from(adata: AssocData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(i64::build_from(adata, stmt, index)? as usize)
    }
}

impl Datum for isize {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to(&self, _stmt: &mut StatementContext, _index: i32) {
        todo!()
    }

    fn build_from(adata: AssocData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(i64::build_from(adata, stmt, index)? as isize)
    }
}

impl Datum for u64 {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to(&self, _stmt: &mut StatementContext, _index: i32) {
        todo!()
    }

    fn build_from(adata: AssocData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(i64::build_from(adata, stmt, index)? as u64)
    }
}

impl Datum for i64 {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        stmt.bind(index, *self).expect("couldn't bind i64")
    }

    fn build_from(_: AssocData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        let val = stmt.read(*index)?;
        *index += 1;
        Ok(val)
    }
}

impl<T: Datum> Datum for Option<T> {
    fn sql_type() -> &'static str {
        T::sql_type()
    }

    fn bind_to(&self, _stmt: &mut StatementContext, _index: i32) {
        todo!()
    }

    fn build_from(_: AssocData, _stmt: &mut StatementRow, _index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        // Ok((stmt.read::<i64, _>(index)? as u64, index+1))
        todo!()
    }
}

impl Datum for bool {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        stmt.bind(index, if *self { 1i64 } else { 0i64 })
            .expect("couldn't bind bool");
    }

    fn build_from(adata: AssocData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(i64::build_from(adata, stmt, index)? != 0)
    }
}

impl Datum for Vec<u8> {
    fn sql_type() -> &'static str {
        "blob"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        stmt.bind(index, self.as_slice())
            .expect("couldn't bind Vec<u8>");
    }

    fn build_from(_: AssocData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        let val = stmt.read(*index)?;
        *index += 1;
        Ok(val)
    }
}

impl<'l, T: Datum> Datum for &'l T {
    fn sql_type() -> &'static str {
        T::sql_type()
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        T::bind_to(self, stmt, index)
    }

    fn build_from(_adata: AssocData, _stmt: &mut StatementRow, _index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        unreachable!()
    }

    fn accept_discriminator(d: &mut impl crate::schema::DatumDiscriminator)
    where
        Self: Sized,
    {
        T::accept_discriminator(d)
    }

    fn accept_entity_visitor(v: &mut impl crate::schema::entity::EntityVisitor) {
        T::accept_entity_visitor(v)
    }
}
